#ifndef STM32F4
  #include <stm32f4xx.h>
  #define STM32F4
#endif

#ifndef STDBOOL
  #include <stdbool.h>
  #define STDBOOL
#endif

#ifndef INITS
  #include "inits.h"
  #define INITS
#endif

#ifndef COM
  #include "com.h"
  #define COM
#endif

/*Initialize Pins:
======== GPIOs ========
Buttons:
PB5   GPIO_IN PullDown -> Button 1
PB6   GPIO_IN PullDown -> Button 2
PB7   GPIO_IN PullDown -> Button 3
PB12  GPIO_IN PullDown -> Button 4
PB13  GPIO_IN PullDown -> Button 5

LEDs:
PE0   GPIO_OUT PushPull -> LED rt Fatal Error
PE1   GPIO_OUT PushPull -> LED rt F-Alert
PE2   GPIO_OUT PushPull -> LED ge Error
PE3   GPIO_OUT PushPull -> LED ge Testmode
PE4   GPIO_OUT PushPull -> LED ge Race
PE5   GPIO_OUT PushPull -> LED gr Fuelpump On
PE6   GPIO_OUT PushPull -> LED gr Throttle On

LCD:
PC2   GPIO_OUT PushPull -> LCD_RS
PC3   GPIO_OUT PushPull -> SPI_Select

Sensor fail signals:
PB0   GPIO_IN PullUp -> Fail APPS-1
PB1   GPIO_IN PullUp -> Fail APPS-2
PB13  GPIO_IN PullUp -> Fail APPS-3
PD0   GPIO_IN PullUp -> Fail TPS-1
PD1   GPIO_IN PullUp -> Fail TPS-2
PD13  GPIO_IN PullUp -> Fail TPS-3
PD14  GPIO_IN PullUp -> Fail BSE-1
PD15  GPIO_IN PullUp -> Fail BSE-2

Sensor signals:
PA2   ADC3_CH02 -> TPS-3
PA3   ADC3_CH03 -> APPS-3
PA4   ADC2_CH04 -> BSE-2
PA5   ADC2_CH05 -> TPS-2
PA6   ADC2_CH06 -> APPS-2
PA7   ADC1_CH07 -> BSE-1
PC4   ADC1_CH14 -> TPS-1
PC5   ADC1_CH15 -> APPS-1

Motec signals:
PA9   Tim1_Ch2
PA10  EXTI10_1

Other:
PE9   GPIO_OUT PushPull -> Fuelpump on
PA0   GPIO_OUT PushPull -> Servo on

======== SPI2 ========
PB10  SPI_SCK
PB14  SPI_MISO
PB15  SPI_MOSI

======== SD-Slot ========
PD2   SDIO_CMD
PC12  SDIO_CK
PC8   SDIO_D0
PC9   SDIO_D1
PC10  SDIO_D2
PC11  SDIO_D3

======== CAN1 ========
PA11  CAN1_RX
PA12  CAN1_TX

======== USART3 ========
PD9   USART3_RX
PD8   USART3_TX
PD11  USART3_CTS
PD12  USART3_RTS
*/

//Initialize clock config
//Siehe: https://github.com/jkerdels/stm32edu/blob/master/src/rcc.c
void initClock(){
  //Enable HSI-clock -> 16MHz
  RCC -> CR |= RCC_CR_HSION;

  //Wait for HSI is stable running
  while(RCC -> CR & !RCC_CR_HSIRDY ){}
  
  //Switch SYSCLK to HSI, SW -> 00
  RCC -> CFGR &= ~RCC_CFGR_SW_0;
  RCC -> CFGR &= ~RCC_CFGR_SW_1;

  //Wait for switching SYSCLK is ready
  while((RCC -> CFGR & RCC_CFGR_SWS_0 == false) && (RCC -> CFGR & RCC_CFGR_SWS_1 == false)){}

  //Config waitstates for flash memory, cpu should run with nearly 180MHz@3,3V -> 5ws after table, page 64
  FLASH -> ACR |= FLASH_ACR_LATENCY_6WS;

  //Config realtime clock, set LSI as clocksource
  RCC -> CSR |= RCC_CSR_LSION;

  //Set AHB clock prescaler for main bus matrix, psc=1 for running with maximum speed -> 180MHz, Presc. =1
  RCC -> CFGR &= ~RCC_CFGR_HPRE_3 & ~RCC_CFGR_HPRE_2 & ~RCC_CFGR_HPRE_1 & ~RCC_CFGR_HPRE_0;

  //Set clock for APB1 to AHB/4 => 180MHz/4 -> 45MHz
  RCC -> CFGR |= RCC_CFGR_PPRE1_DIV8;

  //Set clock for APB2 to AHB/2 => 180MHz/2 -> 90MHz
  RCC -> CFGR |= RCC_CFGR_PPRE2_DIV8;

  //Set phase lock loop
  //Disable PLL for config
  RCC -> CR &= ~RCC_CR_PLLON;
  RCC -> CR &= ~RCC_CR_PLLI2SON;

  //Select HSI for pll clock source
  RCC -> PLLCFGR |= RCC_PLLCFGR_PLLSRC_HSI;

  //Set prescaler for pll clock input: HSI(16MHz) -> input should be 1-2MHz -> Prescaler =10, clock is set to 1,6MHz
  RCC -> PLLCFGR |= RCC_PLLCFGR_PLLM_3 | RCC_PLLCFGR_PLLM_1;

  //Clock = fin*(N/P) -> 1,6MHz*(225/2) for 180MHz
  //Set N=225 -> 0b1110 0001
  RCC -> PLLCFGR |= RCC_PLLCFGR_PLLN_7 | RCC_PLLCFGR_PLLN_6 | RCC_PLLCFGR_PLLN_5 | RCC_PLLCFGR_PLLN_0;
  //Set P=2 0b00
  RCC -> PLLCFGR &= ~RCC_PLLCFGR_PLLP_1 & ~RCC_PLLCFGR_PLLP_0;

  //Enable PLL-Modul
  RCC -> CR |= RCC_CR_PLLON;

  //Wait for PLL is stable running
  while(RCC -> CR & !RCC_CR_PLLRDY){}

  //Switch SYSCLK to PLL_P, SW -> 10
  RCC -> CFGR |= RCC_CFGR_SW_1 & ~RCC_CFGR_SW_0;

  //Wait for clock switch is ready -> SWS = 10
  while((RCC -> CFGR & RCC_CFGR_SWS_PLL) != RCC_CFGR_SWS_PLL){};

}

//Initialize nested Vector Interrupt
void initNVIC(){
  //Enable interrupt request for timer6
  NVIC_EnableIRQ(TIM6_DAC_IRQn);

  //Enalble interrupt request for usart3
  NVIC_EnableIRQ(USART3_IRQn);
}

//Initialize Pinconfiguration
void initPins(){
  //Init Buttons
  RCC -> AHB1ENR |= RCC_AHB1ENR_GPIOBEN;
  BUTTON1_PORT -> MODER |= GPIO_MODER_MODER5_0;
  BUTTON1_PORT -> PUPDR |= GPIO_PUPDR_PUPD5_1;

  RCC -> AHB1ENR |= RCC_AHB1ENR_GPIOBEN;
  BUTTON2_PORT -> MODER |= GPIO_MODER_MODER6_0;
  BUTTON2_PORT -> PUPDR |= GPIO_PUPDR_PUPD6_1;

  RCC -> AHB1ENR |= RCC_AHB1ENR_GPIOBEN;
  BUTTON3_PORT -> MODER |= GPIO_MODER_MODER7_0;
  BUTTON3_PORT -> PUPDR |= GPIO_PUPDR_PUPD7_1;

  RCC -> AHB1ENR |= RCC_AHB1ENR_GPIOBEN;
  BUTTON4_PORT -> MODER |= GPIO_MODER_MODER12_0;
  BUTTON4_PORT -> PUPDR |= GPIO_PUPDR_PUPD12_1;

  RCC -> AHB1ENR |= RCC_AHB1ENR_GPIOBEN;
  BUTTON5_PORT -> MODER |= GPIO_MODER_MODER13_0;
  BUTTON5_PORT -> PUPDR |= GPIO_PUPDR_PUPD13_1;

  //Init LEDs
  RCC -> AHB1ENR |= RCC_AHB1ENR_GPIOEEN;
  LED_FATERR_PORT -> MODER |= GPIO_MODER_MODER0_0;
  LED_FATERR_PORT -> BSRR = LED_FATERR_OFF;

  RCC -> AHB1ENR |= RCC_AHB1ENR_GPIOEEN;
  LED_FATALERT_PORT -> MODER |= GPIO_MODER_MODER1_0;
  LED_FATALERT_PORT -> BSRR = LED_FATALERT_OFF;

  RCC -> AHB1ENR |= RCC_AHB1ENR_GPIOEEN;
  LED_ERR_PORT -> MODER |= GPIO_MODER_MODER2_0;
  LED_ERR_PORT -> BSRR = LED_ERR_OFF;

  RCC -> AHB1ENR |= RCC_AHB1ENR_GPIOEEN;
  LED_TEST_PORT -> MODER |= GPIO_MODER_MODER3_0;
  LED_TEST_PORT -> BSRR = LED_TEST_OFF;

  RCC -> AHB1ENR |= RCC_AHB1ENR_GPIOEEN;
  LED_RACE_PORT -> MODER |= GPIO_MODER_MODER4_0;
  LED_RACE_PORT -> BSRR = LED_RACE_OFF;

  RCC -> AHB1ENR |= RCC_AHB1ENR_GPIOEEN;
  LED_FUELON_PORT -> MODER |= GPIO_MODER_MODER5_0;
  LED_FUELON_PORT -> BSRR = LED_FUELON_OFF;

  RCC -> AHB1ENR |= RCC_AHB1ENR_GPIOEEN;
  LED_THROTTON_PORT -> MODER |= GPIO_MODER_MODER6_0;
  LED_THROTTON_PORT -> BSRR = LED_THROTTON_OFF;

  //Init pins for usart connection
  //D8 => TX
  RCC -> AHB1ENR |= RCC_AHB1ENR_GPIODEN;
  COM_TX_PORT -> MODER |= GPIO_MODER_MODER8_1;
  COM_TX_PORT -> PUPDR |= GPIO_PUPDR_PUPD8_0;
  COM_TX_PORT -> AFR[1] |= GPIO_AFRH_AFSEL8_2 | GPIO_AFRH_AFSEL8_1 | GPIO_AFRH_AFSEL8_0;
  //D9 => RX
  RCC -> AHB1ENR |= RCC_AHB1ENR_GPIODEN;
  COM_RX_PORT -> MODER |= GPIO_MODER_MODER9_1;
  //COM_RX_PORT -> PUPDR |= GPIO_PUPDR_PUPD9_0;
  COM_RX_PORT -> AFR[1] |= GPIO_AFRH_AFSEL9_2 | GPIO_AFRH_AFSEL9_1 | GPIO_AFRH_AFSEL9_0;
  //D11 => CTS
  RCC -> AHB1ENR |= RCC_AHB1ENR_GPIODEN;
  COM_RTS_PORT -> MODER |= GPIO_MODER_MODER12_1;
  COM_RTS_PORT -> PUPDR |= GPIO_PUPDR_PUPD12_0;
  COM_RTS_PORT -> AFR[1] |= GPIO_AFRH_AFSEL11_2 | GPIO_AFRH_AFSEL11_1 | GPIO_AFRH_AFSEL11_0;
  //D12 => RTS
  RCC -> AHB1ENR |= RCC_AHB1ENR_GPIOEEN;
  COM_RTS_PORT -> MODER |= GPIO_MODER_MODER10_1;
  //COM_RTS_PORT -> PUPDR |= GPIO_PUPDR_PUPD10_0;
  COM_RTS_PORT -> AFR[1] |= GPIO_AFRH_AFSEL12_2 | GPIO_AFRH_AFSEL10_1 | GPIO_AFRH_AFSEL12_0;
}

void initTimer(){
  //Initialize timer6 for check user buttons per periodic interrupt
  //ans set LEDs for user interface
  //Tim6 runs @APB1 with 45MHz -> interrupt time 1ms
  RCC -> APB1ENR |= RCC_APB1ENR_TIM6EN;
  TIM6 -> DIER |= TIM_DIER_UIE;
  TIM6 -> PSC = 90;
  TIM6 -> ARR = 500;
}

void initCOM(){
  //Initialize USART for comunication
  //Enable USART3 clock
  RCC -> APB1ENR |= RCC_APB1ENR_USART3EN;
  //Set baudrate 45MHz/48,828125 => 115200Bd with 8x oversampling
  COM -> BRR |= (0x01 << USART_BRR_DIV_Mantissa_Pos); //(0x14 << USART_BRR_DIV_Mantissa_Pos);   //0x14 => 20
  COM -> BRR |= 0b0000;   //0b0000;

  //Set oversampling to 8x-oversampling
  COM -> CR1 |= USART_CR1_OVER8;
  
  //Set wordlenght databits to 8
  COM -> CR1 &= ~USART_CR1_M;

  //Disable parity control
  COM -> CR1 &= ~USART_CR1_PCE;

  //Select parity mode
  COM -> CR1 &= ~USART_CR1_PS;

  //Enable interrupt if TX is empty or RX is'nt empty
  //COM -> CR1 |= USART_CR1_TXEIE | USART_CR1_RXNEIE;

  //Enable transmitter and reciver
  COM -> CR1 |= USART_CR1_TE | USART_CR1_RE;

  //Select stopbits => 1 stopbit
  COM -> CR2 |= ~USART_CR2_STOP_0 & ~USART_CR2_STOP_1;

  //Enable CTS/RTS
  COM -> CR3 |= USART_CR3_CTSE | USART_CR3_RTSE;  
}